@@ -60,6 +60,9 @@ struct NdQueuePrivate
         NotifyScreen  *screen;
 
         guint          update_id;
+
+        GdkDevice     *grabbed_pointer;
+        GdkDevice     *grabbed_keyboard;
 };
 
 enum {
@@ -69,8 +72,7 @@ enum {
 
 static guint signals [LAST_SIGNAL] = { 0, };
 
-static void     nd_queue_class_init     (NdQueueClass   *klass);
-static void     nd_queue_init           (NdQueue        *queue);
+static void     nd_queue_dispose        (GObject        *object);
 static void     nd_queue_finalize       (GObject        *object);
 static void     queue_update            (NdQueue        *queue);
 static void     on_notification_close   (NdNotification *notification,
@@ -224,6 +226,7 @@ nd_queue_class_init (NdQueueClass *klass)
 {
         GObjectClass   *object_class = G_OBJECT_CLASS (klass);
 
+        object_class->dispose = nd_queue_dispose;
         object_class->finalize = nd_queue_finalize;
 
         signals [CHANGED] =
@@ -241,36 +244,37 @@ nd_queue_class_init (NdQueueClass *klass)
 }
 
 static void
-popdown_dock (NdQueue *queue)
+ungrab (NdQueue *queue,
+        guint    time)
 {
-        GdkDisplay *display;
+        if (queue->priv->grabbed_pointer != NULL) {
+                gdk_device_ungrab (queue->priv->grabbed_pointer, time);
+                g_clear_object (&queue->priv->grabbed_pointer);
+        }
+
+        if (queue->priv->grabbed_keyboard != NULL) {
+                gdk_device_ungrab (queue->priv->grabbed_keyboard, time);
+                g_clear_object (&queue->priv->grabbed_keyboard);
+        }
 
-        /* ungrab focus */
-        display = gtk_widget_get_display (queue->priv->dock);
-        gdk_display_keyboard_ungrab (display, GDK_CURRENT_TIME);
-        gdk_display_pointer_ungrab (display, GDK_CURRENT_TIME);
         gtk_grab_remove (queue->priv->dock);
 
         /* hide again */
         gtk_widget_hide (queue->priv->dock);
+}
 
+static void
+popdown_dock (NdQueue *queue)
+{
+        ungrab (queue, GDK_CURRENT_TIME);
         queue_update (queue);
 }
 
 static void
-release_grab (GtkWidget      *widget,
+release_grab (NdQueue        *queue,
               GdkEventButton *event)
 {
-        GdkDisplay *display;
-
-        /* ungrab focus */
-        display = gtk_widget_get_display (widget);
-        gdk_display_keyboard_ungrab (display, event->time);
-        gdk_display_pointer_ungrab (display, event->time);
-        gtk_grab_remove (widget);
-
-        /* hide again */
-        gtk_widget_hide (widget);
+        ungrab (queue, event->time);
 }
 
 /* This is called when the grab is broken for
@@ -392,7 +396,7 @@ on_dock_button_press (GtkWidget      *widget,
         event_widget = gtk_get_event_widget ((GdkEvent *)event);
         g_debug ("Button press: %p dock=%p", event_widget, widget);
         if (event_widget == widget) {
-                release_grab (widget, event);
+                release_grab (queue, event);
                 return TRUE;
         }
 
@@ -466,6 +470,9 @@ nd_queue_init (NdQueue *queue)
         queue->priv->queue = g_queue_new ();
         queue->priv->status_icon = NULL;
 
+        queue->priv->grabbed_pointer = NULL;
+        queue->priv->grabbed_keyboard = NULL;
+
         create_dock (queue);
         create_screen (queue);
 }
@@ -498,6 +505,18 @@ destroy_screen (NdQueue *queue)
         queue->priv->screen = NULL;
 }
 
+static void
+nd_queue_dispose (GObject *object)
+{
+        NdQueue *queue;
+
+        queue = ND_QUEUE (object);
+
+        g_clear_object (&queue->priv->grabbed_pointer);
+        g_clear_object (&queue->priv->grabbed_keyboard);
+
+        G_OBJECT_CLASS (nd_queue_parent_class)->dispose (object);
+}
 
 static void
 nd_queue_finalize (GObject *object)
@@ -551,15 +570,18 @@ nd_queue_length (NdQueue *queue)
 static NdStack *
 get_stack_with_pointer (NdQueue *queue)
 {
+        GdkDisplay *display;
+        GdkDeviceManager *device_manager;
+        GdkDevice *pointer;
         GdkScreen *screen;
         int        x, y;
         int        monitor_num;
 
-        gdk_display_get_pointer (gdk_display_get_default (),
-                                 &screen,
-                                 &x,
-                                 &y,
-                                 NULL);
+        display = gdk_display_get_default ();
+        device_manager = gdk_display_get_device_manager (display);
+        pointer = gdk_device_manager_get_client_pointer (device_manager);
+
+        gdk_device_get_position (pointer, &screen, &x, &y);
         monitor_num = gdk_screen_get_monitor_at_point (screen, x, y);
 
         if (monitor_num >= queue->priv->screen->n_stacks) {
@@ -658,6 +680,8 @@ update_dock (NdQueue *queue)
         int          monitor_num;
         GdkScreen   *screen;
         GdkRectangle area;
+        GtkStatusIcon *status_icon;
+        gboolean visible;
 
         g_return_if_fail (queue);
 
@@ -666,8 +690,9 @@ update_dock (NdQueue *queue)
                 gtk_container_remove (GTK_CONTAINER (queue->priv->dock_scrolled_window), child);
 
         child = gtk_box_new (GTK_ORIENTATION_VERTICAL, 6);
-        gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW (queue->priv->dock_scrolled_window),
-                                               child);
+        gtk_container_add (GTK_CONTAINER (queue->priv->dock_scrolled_window),
+                           child);
+
         gtk_container_set_focus_hadjustment (GTK_CONTAINER (child),
                                              gtk_scrolled_window_get_hadjustment (GTK_SCROLLED_WINDOW (queue->priv->dock_scrolled_window)));
         gtk_container_set_focus_vadjustment (GTK_CONTAINER (child),
@@ -685,21 +710,28 @@ update_dock (NdQueue *queue)
                 gtk_widget_show (GTK_WIDGET (box));
                 gtk_box_pack_start (GTK_BOX (child), GTK_WIDGET (box), FALSE, FALSE, 0);
 
-                sep = gtk_hseparator_new ();
+                sep = gtk_separator_new (GTK_ORIENTATION_HORIZONTAL);
                 gtk_widget_show (sep);
                 gtk_box_pack_start (GTK_BOX (child), sep, FALSE, FALSE, 0);
         }
         gtk_widget_show (child);
 
-        if (queue->priv->status_icon != NULL
-            && gtk_status_icon_get_visible (GTK_STATUS_ICON (queue->priv->status_icon))) {
-                gtk_widget_get_preferred_height (child,
-                                                 &min_height,
-                                                 &height);
-                gtk_status_icon_get_geometry (GTK_STATUS_ICON (queue->priv->status_icon),
-                                              &screen,
-                                              &area,
-                                              NULL);
+        status_icon = queue->priv->status_icon;
+        visible = FALSE;
+
+        if (status_icon != NULL) {
+                G_GNUC_BEGIN_IGNORE_DEPRECATIONS
+                visible = gtk_status_icon_get_visible (status_icon);
+                G_GNUC_END_IGNORE_DEPRECATIONS
+        }
+
+        if (visible) {
+                gtk_widget_get_preferred_height (child, &min_height, &height);
+
+                G_GNUC_BEGIN_IGNORE_DEPRECATIONS
+                gtk_status_icon_get_geometry (status_icon, &screen, &area, NULL);
+                G_GNUC_END_IGNORE_DEPRECATIONS
+
                 monitor_num = gdk_screen_get_monitor_at_point (screen, area.x, area.y);
                 gdk_screen_get_monitor_geometry (screen, monitor_num, &area);
                 height = MIN (height, (area.height / 2));
@@ -725,13 +757,21 @@ popup_dock (NdQueue *queue,
         int            monitor_num;
         GdkRectangle   monitor;
         GtkRequisition dock_req;
+        GtkStatusIcon *status_icon;
+        GdkWindow *window;
+        GdkDeviceManager *device_manager;
+        GList *list;
+        GList *link;
+        gboolean grabbed;
 
         update_dock (queue);
 
-        res = gtk_status_icon_get_geometry (GTK_STATUS_ICON (queue->priv->status_icon),
-                                            &screen,
-                                            &area,
-                                            &orientation);
+        status_icon = queue->priv->status_icon;
+
+        G_GNUC_BEGIN_IGNORE_DEPRECATIONS
+        res = gtk_status_icon_get_geometry (status_icon, &screen, &area, &orientation);
+        G_GNUC_END_IGNORE_DEPRECATIONS
+
         if (! res) {
                 g_warning ("Unable to determine geometry of status icon");
                 return FALSE;
@@ -783,21 +823,52 @@ popup_dock (NdQueue *queue,
         /* grab focus */
         gtk_grab_add (queue->priv->dock);
 
-        if (gdk_pointer_grab (gtk_widget_get_window (queue->priv->dock), TRUE,
-                              GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
-                              GDK_POINTER_MOTION_MASK | GDK_SCROLL_MASK, NULL, NULL,
-                              time)
-            != GDK_GRAB_SUCCESS) {
-                gtk_grab_remove (queue->priv->dock);
-                gtk_widget_hide (queue->priv->dock);
+        display = gtk_widget_get_display (queue->priv->dock);
+        window = gtk_widget_get_window (queue->priv->dock);
+        device_manager = gdk_display_get_device_manager (display);
+
+        grabbed = FALSE;
+        list = gdk_device_manager_list_devices (device_manager, GDK_DEVICE_TYPE_MASTER);
+        for (link = list; link != NULL; link = g_list_next (link)) {
+                GdkDevice *device = GDK_DEVICE (link->data);
+                if (gdk_device_get_source (device) == GDK_SOURCE_KEYBOARD)
+                        continue;
+                if (gdk_device_grab (device, window,
+                                     GDK_OWNERSHIP_NONE, TRUE,
+                                     GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
+                                     GDK_POINTER_MOTION_MASK | GDK_SCROLL_MASK,
+                                     NULL, time) == GDK_GRAB_SUCCESS) {
+                        queue->priv->grabbed_pointer = g_object_ref (device);
+                        grabbed = TRUE;
+                        break;
+                }
+        }
+        g_list_free (list);
+
+        if (grabbed == FALSE) {
+                ungrab (queue, time);
                 return FALSE;
         }
 
-        if (gdk_keyboard_grab (gtk_widget_get_window (queue->priv->dock), TRUE, time) != GDK_GRAB_SUCCESS) {
-                display = gtk_widget_get_display (queue->priv->dock);
-                gdk_display_pointer_ungrab (display, time);
-                gtk_grab_remove (queue->priv->dock);
-                gtk_widget_hide (queue->priv->dock);
+        grabbed = FALSE;
+        list = gdk_device_manager_list_devices (device_manager, GDK_DEVICE_TYPE_MASTER);
+        for (link = list; link != NULL; link = g_list_next (link)) {
+                GdkDevice *device = GDK_DEVICE (link->data);
+                if (gdk_device_get_source (device) != GDK_SOURCE_KEYBOARD)
+                        continue;
+                if (gdk_device_grab (device, window,
+                                     GDK_OWNERSHIP_NONE, TRUE,
+                                     GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK,
+                                     NULL, time) == GDK_GRAB_SUCCESS) {
+                        queue->priv->grabbed_keyboard = g_object_ref (device);
+                        grabbed = TRUE;
+                        break;
+                }
+        }
+        g_list_free (list);
+
+        if (grabbed == FALSE) {
+                ungrab (queue, time);
                 return FALSE;
         }
 
@@ -862,9 +933,12 @@ update_idle (NdQueue *queue)
                 }
 
                 if (queue->priv->status_icon == NULL) {
+                        G_GNUC_BEGIN_IGNORE_DEPRECATIONS
                         queue->priv->status_icon = gtk_status_icon_new ();
-                        gtk_status_icon_set_title (GTK_STATUS_ICON (queue->priv->status_icon),
+                        gtk_status_icon_set_title (queue->priv->status_icon,
                                                    _("Notifications"));
+                        G_GNUC_END_IGNORE_DEPRECATIONS
+
                         g_signal_connect (queue->priv->status_icon,
                                           "activate",
                                           G_CALLBACK (on_status_icon_activate),
@@ -883,13 +957,20 @@ update_idle (NdQueue *queue)
                         GIcon *icon;
                         /* FIXME: use a more appropriate icon here */
                         icon = g_themed_icon_new ("mail-message-new");
+
+                        G_GNUC_BEGIN_IGNORE_DEPRECATIONS
                         queue->priv->numerable_icon = gtk_numerable_icon_new (icon);
+                        G_GNUC_END_IGNORE_DEPRECATIONS
+
                         g_object_unref (icon);
                 }
+
+                G_GNUC_BEGIN_IGNORE_DEPRECATIONS
                 gtk_numerable_icon_set_count (GTK_NUMERABLE_ICON (queue->priv->numerable_icon), num);
                 gtk_status_icon_set_from_gicon (queue->priv->status_icon,
                                                 queue->priv->numerable_icon);
                 gtk_status_icon_set_visible (queue->priv->status_icon, TRUE);
+                G_GNUC_END_IGNORE_DEPRECATIONS
 
                 maybe_show_notification (queue);
         } else {
